#include <stdio.h>
#include <zneo.h>
#include "zneo2.h"

////////////////////////////////////////
//
// Free for private and commercial use.
//
// 2008-11-06
//
////////////////////////////////////////

////////////////////////////////////////
// General-Purpose Input/Output (GPIO)

typedef struct
{
	volatile uint8_t near*	m_pucPxIN;		// Port A-K Input Data Registers (PxIN)
	volatile uint8_t near*	m_pucPxOUT;		// Port A-K Output Data Registers (PxOUT)
	volatile uint8_t near*	m_pucPxDD;		// Port A-K Data Direction Registers (PxDD)
	volatile uint8_t near*	m_pucPxHDE;		// Port A-K High Drive Enable Registers (PxHDE)
	volatile uint16_t near*	m_pusPxAF;		// Port A-K Alternate Function Registers (PxAF)
	volatile uint8_t near*	m_pucPxAFL;		// Port A-K Alternate Function Low Registers (PxAFL)
	volatile uint8_t near*	m_pucPxOC;		// Port A-K Output Control Registers (PxOC)
	volatile uint8_t near*	m_pucPxPUE;		// Port A-K Pull-Up Enable Registers (PxPUE)
	volatile uint8_t near*	m_pucPxSMRE;	// Port A-K Stop Mode Recovery Source Enable Registers (PxSMRE)
}
GPIO_REG;

static const GPIO_REG GpioReg[] =
{
	{ &PAIN, &PAOUT, &PADD, &PAHDE, &PAAF, NULL, &PAOC, &PAPUE, &PASMRE },
	{ &PBIN, &PBOUT, &PBDD, &PBHDE, NULL, &PBAFL, &PBOC, &PBPUE, &PBSMRE },
	{ &PCIN, &PCOUT, &PCDD, &PCHDE, &PCAF, NULL, &PCOC, &PCPUE, &PCSMRE },
	{ &PDIN, &PDOUT, &PDDD, &PDHDE, &PDAF, NULL, &PDOC, &PDPUE, &PDSMRE },
	{ &PEIN, &PEOUT, &PEDD, &PEHDE, NULL, NULL, &PEOC, &PEPUE, &PESMRE },
	{ &PFIN, &PFOUT, &PFDD, &PFHDE, NULL, &PFAFL, &PFOC, &PFPUE, &PFSMRE },
	{ &PGIN, &PGOUT, &PGDD, &PGHDE, NULL, &PGAFL, &PGOC, &PGPUE, &PGSMRE },
	{ &PHIN, &PHOUT, &PHDD, &PHHDE, &PHAF, NULL, &PHOC, &PHPUE, &PHSMRE },

#ifdef __ZNEO_PORTJ
	{ &PJIN, &PJOUT, &PJDD, &PJHDE, NULL, NULL, &PJOC, &PJPUE, &PJSMRE },
#else
	{ NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL },
#endif	// __ZNEO_PORTJ

#ifdef __ZNEO_PORTK
	{ &PKIN, &PKOUT, &PKDD, &PKHDE, NULL, &PKAFL, &PKOC, &PKPUE, &PKSMRE },
#else
	{ NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL },
#endif	// __ZNEO_PORTK
};

static const GPIO_REG* GetGPIO(GPIO cGPIO)
{
#ifdef NDEBUG
	return &GpioReg[cGPIO >> 3];
#else
	return cGPIO >= PORTA0 && cGPIO <= PORT_UPPER ? &GpioReg[cGPIO >> 3] : NULL;
#endif
}

////////////////////////////////////////

// Set Input/Output Mode
void SetIOMode(GPIO cGPIO, uint8_t ucAltFunc, BOOL bOutput, BOOL bHighCurrent, BOOL bOpenDrain, BOOL bPullUp, BOOL bStopModeRecovery)
{
	uint8_t ucBitMask = _BV(cGPIO & 0x07);
	const GPIO_REG* pGpioReg = GetGPIO(cGPIO);
#ifndef NDEBUG
	if (!pGpioReg)
		return;
#endif
	if (pGpioReg->m_pusPxAF)
	{
		uint16_t usAltMask = ((uint16_t)(ucAltFunc & 0x02) << 7) | (ucAltFunc & 0x01);

		*pGpioReg->m_pusPxAF &= ~(((uint16_t)ucBitMask << 8) | ucBitMask);
		*pGpioReg->m_pusPxAF |= usAltMask << (cGPIO & 0x07);
	}
	else if (pGpioReg->m_pucPxAFL)
	{
		uint8_t ucAltMask = ucAltFunc & 0x01;

		*pGpioReg->m_pucPxAFL &= ~ucBitMask;
		*pGpioReg->m_pucPxAFL |= ucAltMask << (cGPIO & 0x07);
	}

	if (pGpioReg->m_pucPxDD)
	{
		if (!bOutput)
			*pGpioReg->m_pucPxDD |= ucBitMask;
		else
			*pGpioReg->m_pucPxDD &= ~ucBitMask;
	}

	if (pGpioReg->m_pucPxHDE)
	{
		if (bHighCurrent)
			*pGpioReg->m_pucPxHDE |= ucBitMask;
		else
			*pGpioReg->m_pucPxHDE &= ~ucBitMask;
	}

	if (pGpioReg->m_pucPxOC)
	{
		if (bOpenDrain)
			*pGpioReg->m_pucPxOC |= ucBitMask;
		else
			*pGpioReg->m_pucPxOC &= ~ucBitMask;
	}

	if (pGpioReg->m_pucPxPUE)
	{
		if (bPullUp)
			*pGpioReg->m_pucPxPUE |= ucBitMask;
		else
			*pGpioReg->m_pucPxPUE &= ~ucBitMask;
	}

	if (pGpioReg->m_pucPxSMRE)
	{
		if (bStopModeRecovery)
			*pGpioReg->m_pucPxSMRE |= ucBitMask;
		else
			*pGpioReg->m_pucPxSMRE &= ~ucBitMask;
	}
}

// Read one Port Bit
BOOL PGet(GPIO cGPIO)
{
	const GPIO_REG* pGpioReg = GetGPIO(cGPIO);
#ifndef NDEBUG
	if (!pGpioReg || !pGpioReg->m_pucPxIN)
		return FALSE;
#endif
	return *pGpioReg->m_pucPxIN & _BV(cGPIO & 0x07) ? TRUE : FALSE;
}

// Write one Port Bit
void PSet(GPIO cGPIO, BOOL bSet)
{
	const GPIO_REG* pGpioReg = GetGPIO(cGPIO);
#ifndef NDEBUG
	if (!pGpioReg || !pGpioReg->m_pucPxOUT)
		return;
#endif
	if (bSet)
		*pGpioReg->m_pucPxOUT |= _BV(cGPIO & 0x07);
	else
		*pGpioReg->m_pucPxOUT &= ~_BV(cGPIO & 0x07);
}

////////////////////////////////////////
// The Delay1ms() functions perform a delay of uTime milliseconds

void Delay1ms(uint32_t uTime)
{
asm("				JP   Delay1ms_3					");
asm("	Delay1ms_1:	LD   R0,#(SYSTEM_FREQ / 14000)	");
asm("	Delay1ms_2:	ADD  R0,#-1						");
asm("				EXT.SW R2,R0					");
asm("				JP   nz,Delay1ms_2				");
asm("	Delay1ms_3:	LD   R0,R1						");
asm("				ADD  R1,#-1						");
asm("				CP   R0,#0						");
asm("				JP   nz,Delay1ms_1				");
}

////////////////////////////////////////
// The _rotl and _rotr functions rotate uValue by iShift bits.
// _rotl rotates the value left. _rotr rotates the value right.
// Both wrap bits rotated off one end of value to the other end.

uint32_t _rotl(uint32_t uValue, int32_t iShift)
{
asm("	RL	R1,R2	");
	return uValue;
}

uint32_t _rotr(uint32_t uValue, int32_t iShift)
{
asm("	NEG	R2		");
asm("	RL	R1,R2	");
	return uValue;
}

////////////////////////////////////////
// The copysign function returns x but with the sign of y

float copysignf(float x, float y)
{
asm("	AND	 R1,#%7FFFFFFF	");
asm("	AND  R2,#%80000000	");
asm("	OR   R1,R2			");
	return x;
}

double copysign(double x, double y)
{
	return copysignf(x, y);
}

////////////////////////////////////////
// The fmax function compares two values and returns the larger one.

float fmaxf(float x, float y)
{
	return max(x, y);
}

double fmax(double x, double y)
{
	return max(x, y);
}

////////////////////////////////////////
// The fmin function compares two values and returns the smaller one.

float fminf(float x, float y)
{
	return min(x, y);
}

double fmin(double x, double y)
{
	return min(x, y);
}

////////////////////////////////////////
